/*
 *
 *  Copyright (C) 1993-2000, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Healthcare Information and Communication Systems
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Module:  imagectn
 *
 *  Author:  Andrew Hewett
 *
 *  Purpose: Telnet Initiator (ti) Main Program
 *
 *  Last Update:      $Author: meichel $
 *  Update Date:      $Date: 2000/04/14 16:38:03 $
 *  Source File:      $Source: /share/dicom/cvs-depot/dcmtk/imagectn/apps/ti.cc,v $
 *  CVS/RCS Revision: $Revision: 1.22 $
 *  Status:           $State: Exp $
 *
 *  CVS/RCS Log at end of file
 *
 */

#include "osconfig.h"    /* make sure OS specific configuration is included first */

#ifdef HAVE_STDLIB_H
#ifndef  _BCB4
/* workaround for bug in Borland C++ Builder 4 */
BEGIN_EXTERN_C
#endif
#include <stdlib.h>
#ifndef  _BCB4
END_EXTERN_C
#endif
#endif

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#include <time.h>
#include <errno.h>
#include <signal.h>
END_EXTERN_C

#include "ti.h"
#include "assoc.h"
#include "dimse.h"
#include "cnf.h"
#include "dcdict.h"
#include "cmdlnarg.h"
#include "dcdebug.h"
#ifndef  NODISPLAY
#include "dispuser.h"
#endif
#include "tiui.h"
#include "tinet.h"
#include "tiquery.h"
#include "dcuid.h"    /* for dcmtk version name */

static char rcsid[] = "$dcmtk: ti v"
  OFFIS_DCMTK_VERSION " " OFFIS_DCMTK_RELEASEDATE " $";

#define DEFAULTAPPLICATIONTITLE	"TELNET_INITIATOR"	/* default */


char* progname = NULL;
OFBool verbose = OFFalse;
OFBool debug = OFFalse;
static const char *configFileName = "configrc";
OFBool disableDisplay = OFTrue;
OFBool ctnDisplayCmdLineOverride = OFFalse;
OFBool displayTransferProgress = OFTrue;
OFBool disableImageDisplay = OFFalse;

#define MAXREMOTEDBTITLES 20
char *remoteDBTitles[MAXREMOTEDBTITLES];
int remoteDBTitlesCount = 0;

TI_Config conf;


/*
 * My usage stuff
 */ 

static void
shortusage()
{
    fprintf(stderr, "\
usage: %s [-c configrc][-D][-DD][-t title][-v][-d][-I][-P][-R title]\n\
          [-b n][-u] peer ...\n", 
	progname);
}

static void
fullusage()
{
    fprintf(stderr, "%s\n\n", rcsid);
    shortusage();
    fprintf(stderr,
"parameters:\n"
"    peer	peer DPN host name or vendor name (from config file)\n"
"options:\n"
"    -c configrc use specific configuration file (default: %s)\n"
"    -u          disable generation of new VRs (UN/UT)\n"
"    -D		enable display\n"
"    -DD		disable display\n"
"    -t title	my calling AE title (default: %s)\n"
"    -v		verbose mode\n"
"    -d		debug mode\n"
"    -I		disable image display\n"
"    -P		do not display image transfer progress\n"
"    -R title    remote DB AE title\n"
"    -b n	set max receive pdu to n bytes (default: %d)\n",
    configFileName,  DEFAULTAPPLICATIONTITLE, ASC_DEFAULTMAXPDU);
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
	va_list args;
	
	fprintf(stderr, "%s: ", progname);
	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);
	fprintf(stderr, "\n");
}

static OFBool 
addPeerName(TI_Config *aConfig, const char *peerName);
static void
createConfigEntries(TI_Config *aConfig);
static void 
printConfig(TI_Config *aConfig);

#ifdef SIGNAL_HANDLER_WITH_ELLIPSE
static void signalHandler(...);
#else
static void signalHandler(int);
#endif


/*
 * Program Entry Point
 */

int 
main(int argc, char *argv[])
{
    CONDITION cond;
    int i;
    const char **vendorHosts; 
    const char **aeTitleList;
    int maxPDU;

#ifdef HAVE_GUSI_H
    /* needed for Macintosh */
    GUSISetup(GUSIwithSIOUXSockets);
    GUSISetup(GUSIwithInternetSockets);
#endif

#ifdef HAVE_WINSOCK_H
    WSAData winSockData;
    /* we need at least version 1.1 */
    WORD winSockVersionNeeded = MAKEWORD( 1, 1 );
    WSAStartup(winSockVersionNeeded, &winSockData);
#endif

    prepareCmdLineArgs(argc, argv, "echoscu");

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 2) {
	fullusage();
    }

    bzero((char*)&conf, sizeof(conf));
    conf.myAETitle = DEFAULTAPPLICATIONTITLE;
    conf.maxReceivePDULength = -1;
   
    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'D':
	    disableDisplay = (argv[i][2] == 'D');
	    ctnDisplayCmdLineOverride = OFTrue;
#ifdef NODISPLAY
	    fprintf(stderr, "Warning: not compiled for ctndisplay\n");
#endif	    
	    break;
	case 'u':
	    dcmEnableUnknownVRGeneration.set(OFFalse);
	    dcmEnableUnlimitedTextVRGeneration.set(OFFalse);
	    break;
	case 'v':
	    verbose = OFTrue;
	    break;
	case 'd':
	    debug = OFTrue;
	    verbose = OFTrue;
	    break;
	case 'I':
	    disableImageDisplay = OFTrue;
	    break;
	case 'P':
	    displayTransferProgress = OFFalse;
	    break;
	case 'c':
	    if (i++ < argc) {
		configFileName = argv[i];
	    } else
		usage();
	    break;
	case 't':
	    if (i++ < argc) {
		conf.myAETitle = argv[i];
	    } else
		usage();
	    break;
	case 'R':
	    if (i++ < argc) {
	        if (remoteDBTitlesCount < MAXREMOTEDBTITLES) {
		    remoteDBTitles[remoteDBTitlesCount++] = argv[i];
		} else {
		    errmsg("Too many remote DB titles (max: %d)", 
		        MAXREMOTEDBTITLES);
		    errmsg("Ignored temote DB: %s", argv[i]);
		}
	    } else
		usage();
	    break;
	case 'b':
	    if (((i + 1) < argc) &&
		(sscanf(argv[i + 1], "%d", &conf.maxReceivePDULength) == 1)) {
		i++;		/* Maximum Receive PDU Length parsed */
		if (conf.maxReceivePDULength < ASC_MINIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too small",
			   conf.maxReceivePDULength);
		    usage();
		} else if (conf.maxReceivePDULength > ASC_MAXIMUMPDUSIZE) {
		    errmsg("Maximum receive PDU length (%d) too big",
			   conf.maxReceivePDULength);
		    usage();
		}
	    } else {
		usage();
	    }
	    break;
	default:
	    usage();
	}
    }

    if ((argc - i) < 1) {
	errmsg("Peer Host Name missing");
	usage();
    }

    if ((argc - i) > TI_MAXPEERS) {
	errmsg("Too many peers specified (maximum: %d): %d", 
	    TI_MAXPEERS, (argc - i));
	usage();
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    if (access(configFileName, R_OK) < 0) {
	errmsg("cannot access %s: %s", configFileName, 
		strerror(errno));
	usage();
    }

    if (!CNF_init(configFileName)) {
	errmsg("bad config file: %s", configFileName);
	usage();
    }

    if (verbose) {
	CNF_printConfig();    
    }

    maxPDU = CNF_getMaxPDUSize();
    if (maxPDU == 0) {
        /* not set, set default */
	maxPDU = ASC_DEFAULTMAXPDU;
    }
    
    if (maxPDU < ASC_MINIMUMPDUSIZE || maxPDU > ASC_MAXIMUMPDUSIZE) {
        errmsg("bad MaxPDUSize (%d) in config file: %s", 
	    maxPDU, configFileName);
	usage();
    }

    /* if nothing was set on command line then use config file value */
    if (conf.maxReceivePDULength == -1) {
        conf.maxReceivePDULength = maxPDU;
    }


    /* peer to call */
    while (i < argc) {
        int n;
	if ((n = CNF_aeTitlesForSymbolicName(argv[i], &aeTitleList)) > 0) {
	    int j;
	    const char *peerName = NULL;
	    int portNumber;
	    for (j=0; j<n; j++) {
		if (CNF_peerForAETitle(aeTitleList[j], &peerName, &portNumber)) {
		    addPeerName(&conf, peerName);
		}
	    }
	    if (aeTitleList) free(aeTitleList);
	    aeTitleList = NULL;
        } else if ((n = CNF_HostNamesForVendor(argv[i], &vendorHosts)) > 0) {
	    int j;
	    for (j = 0; j<n; j++) {
	        addPeerName(&conf, vendorHosts[j]);
	    }
	    if (vendorHosts) free(vendorHosts); 
	    vendorHosts = NULL;
	} else {
	    addPeerName(&conf, argv[i]);
	}
        i++;
    }
    conf.peerHostName = conf.peerNames[0];	/* default */

    /* load up configuration info */
    createConfigEntries(&conf);

    if (conf.dbCount < 1) {
	errmsg("no accessable databases");
	usage();
    }

    if (verbose) {
	printConfig(&conf);
    }

    /* set debugging if required */
    DUL_Debug(debug);
    DIMSE_debug(debug);
    SetDebugLevel(((debug)?(3):(0)));	/* dcmdata debugging */

    /* make sure data dictionary is loaded */
    if (!dcmDataDict.isDictionaryLoaded()) {
	fprintf(stderr, "Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /* start up network */
    cond = ASC_initializeNetwork(NET_REQUESTOR, 0, 200, &conf.net);
    if (!SUCCESS(cond)) {
	errmsg("Error initialising network:");
	COND_DumpConditions();
	exit(1);
    }

#ifndef NODISPLAY
    /* initialize display */
    if (!ctnDisplayCmdLineOverride) {
	disableDisplay = (!CNF_getDisplay());
    }
    DSU_setDisplayMode(!disableDisplay);
    if (!DSU_initDisplay(DSU_TIUser)) {
	disableDisplay = OFTrue;
    }
#endif

#ifdef SIGHUP
    signal(SIGHUP, signalHandler);
#endif
#ifdef SIGINT
    signal(SIGINT, signalHandler);
#endif
#ifdef SIGQUIT
    signal(SIGQUIT, signalHandler);
#endif
#ifdef SIGTERM
    signal(SIGTERM, signalHandler);
#endif

    /*
     * Do the real work
     */
    TI_userInput(&conf);
    

    /* clean up network */
    cond = ASC_dropNetwork(&conf.net);
    if (!SUCCESS(cond)) {
	errmsg("Error dropping network:");
	COND_DumpConditions();
	exit(1);
    }

#ifdef HAVE_WINSOCK_H
    WSACleanup();
#endif

    return 0;
}

static void 
printConfig(TI_Config *aConfig)
{
    int i,j;

    printf("TI Configuration:\n");
    printf("My AE Title: %s\n", aConfig->myAETitle);
    printf("DatabaseTitles    Peer AE Titles\n");
    for (i=0; i<aConfig->dbCount; i++) {
	printf("%-18s", aConfig->dbEntries[i]->title);
	for (j=0; j<aConfig->dbEntries[i]->peerTitleCount; j++) {
	    printf("%s ", aConfig->dbEntries[i]->peerTitles[j]);
	}
	printf("\n");
    }
}

static OFBool
addPeerName(TI_Config * aConfig, const char *peerName)
{
    int k;
    OFBool found = OFFalse;
    const char **aeTitles;

    /* only hosts not already known */
    for (k = 0; !found && k<aConfig->peerNamesCount; k++) {
	found = (strcmp(aConfig->peerNames[k], peerName) == 0);
    }
    if (found) {
    	/* already there */
        return OFTrue;
    }
    if (CNF_aeTitlesForPeer(peerName, &aeTitles) <= 0) {
        /* no titles defined */
	errmsg("no AE titles defined (in: %s) for peer: %s", configFileName,
	    peerName);
	return OFFalse;
    }
    free(aeTitles);	/* do not need aeTitles */

    aConfig->peerNames[aConfig->peerNamesCount] = (char*)peerName;
    aConfig->peerNamesCount++;

    return OFTrue;
}

static OFBool
findDBPeerTitles(TI_Config * /*aConfig*/, TI_DBEntry *dbEntry, const char *peer)
{
    const char **peerTitles = NULL;
    int peerTitleCount = 0;
    int i;

    /* discover all known AETitles for peer */
    peerTitleCount = CNF_aeTitlesForPeer(peer,
	&peerTitles);
    if (peerTitleCount <= 0) {
	errmsg("no AE titles defined (in: %s) for peer: %s", configFileName,
	    peer);
	return OFFalse;	
    }

    /* check to make sure peer+AEtitle has access to database areas */
    for (i=0; i<peerTitleCount; i++) {
	if (CNF_peerInAETitle(dbEntry->title,
	    peerTitles[i], peer)) {

	    /* add peer title to database's peer title list */
	    if (dbEntry->peerTitles == NULL) {
		dbEntry->peerTitles = (const char**) malloc(sizeof(const char*));
	    } else {
		dbEntry->peerTitles = (const char**) realloc(dbEntry->peerTitles,
		    (dbEntry->peerTitleCount + 1) * sizeof(const char*));
	    }
	    dbEntry->peerTitles[dbEntry->peerTitleCount] = peerTitles[i];
	    dbEntry->peerTitleCount++;
	}
    }

    /* throw away the old list */
    free(peerTitles);

    return (dbEntry->peerTitleCount > 0);
}

static void
createConfigEntries(TI_Config *aConfig)
{
    const char **ctnTitles = NULL;
    int ctnTitleCount = 0;
    int i, j;
    TI_DBEntry *dbEntry = NULL;

    aConfig->dbCount = 0;

    /* discover all the known CTN AE titles */
    ctnTitleCount = CNF_ctnTitles(&ctnTitles);

    for(i=0; i<ctnTitleCount; i++) {
	if (!TI_dbReadable(ctnTitles[i])) {
	    errmsg("Warning: db does not exist: %s", ctnTitles[i]);
	} else {
	    dbEntry = (TI_DBEntry*) malloc(sizeof(TI_DBEntry));
	    bzero((char*)dbEntry, sizeof(*dbEntry));
	    dbEntry->title = ctnTitles[i];

            for (j=0; j<aConfig->peerNamesCount; j++) {
                findDBPeerTitles(aConfig, dbEntry, aConfig->peerNames[j]);
            }

	    if (dbEntry->peerTitleCount > 0) {
	        /* add database to list, it is accessable by something */
	        if (aConfig->dbEntries == NULL) {
	            aConfig->dbEntries = 
		        (TI_DBEntry**) malloc(sizeof(TI_DBEntry*));
	        } else {
	            aConfig->dbEntries = (TI_DBEntry**) realloc(aConfig->dbEntries,
		        (aConfig->dbCount + 1) * sizeof(TI_DBEntry*));
	        }
	        aConfig->dbEntries[aConfig->dbCount] = dbEntry;
	        aConfig->dbCount++;
	    } else {
	        free(dbEntry);
	    }
	}
    }

    /* throw away the old lists */
    free(ctnTitles);

    /* add any remote DB entries */
    for (i=0; i<remoteDBTitlesCount; i++) {
        const char *peerName = NULL;
	int portNumber;
	if (CNF_peerForAETitle(remoteDBTitles[i], &peerName, &portNumber)) {

	    /* add DB */
	    dbEntry = (TI_DBEntry*) malloc(sizeof(TI_DBEntry));
	    bzero((char*)dbEntry, sizeof(*dbEntry));
	    dbEntry->title = remoteDBTitles[i];
	    dbEntry->isRemoteDB = OFTrue;

	    if (aConfig->dbEntries == NULL) {
	        aConfig->dbEntries = 
		    (TI_DBEntry**) malloc(sizeof(TI_DBEntry*));
	    } else {
	        aConfig->dbEntries = (TI_DBEntry**) realloc(aConfig->dbEntries,
		    (aConfig->dbCount + 1) * sizeof(TI_DBEntry*));
	    }
	    aConfig->dbEntries[aConfig->dbCount] = dbEntry;
	    aConfig->dbCount++;

            for (j=0; j<aConfig->peerNamesCount; j++) {
    		const char **peerTitles = NULL;
   		int peerTitleCount = 0;
    		int k;

                peerTitleCount = CNF_aeTitlesForPeer(aConfig->peerNames[j],
		    &peerTitles);
    		if (peerTitleCount <= 0) {
		    errmsg("no AE titles defined (in: %s) for peer: %s", 
		        configFileName, aConfig->peerNames[j]);
    		}

	    	for (k=0; k<peerTitleCount; k++) {
	    	    /* add peer title to database's peer title list */
	    	    if (dbEntry->peerTitles == NULL) {
		    	dbEntry->peerTitles = (const char**) malloc(sizeof(const char*));
	    	    } else {
			dbEntry->peerTitles = (const char**) realloc(dbEntry->peerTitles,
		    	(dbEntry->peerTitleCount + 1) * sizeof(const char*));
	    	    }
	    	    dbEntry->peerTitles[dbEntry->peerTitleCount] = peerTitles[k];
	    	    dbEntry->peerTitleCount++;
	    	}

    		/* throw away the old list */
    		free(peerTitles);

            }	    
	}
    }
}

/*
 * Handle interrupt signals.  
 * We only really need to make sure that the display is clear
 * before quiting.
 */

#ifdef SIGNAL_HANDLER_WITH_ELLIPSE
static void signalHandler(...)
#else
static void signalHandler(int)
#endif
{
    TI_detatchAssociation(&conf, OFTrue);

    exit(1);
}

/*
** CVS Log
** $Log: ti.cc,v $
** Revision 1.22  2000/04/14 16:38:03  meichel
** Global VR generation flags are now derived from OFGlobal and, thus,
**   safe for use in multi-thread applications.
**
** Revision 1.21  2000/03/08 16:41:02  meichel
** Updated copyright header.
**
** Revision 1.20  2000/02/29 11:57:54  meichel
** Removed support for VS value representation. This was proposed in CP 101
**   but never became part of the standard.
**
** Revision 1.19  2000/02/23 15:13:14  meichel
** Corrected macro for Borland C++ Builder 4 workaround.
**
** Revision 1.18  2000/02/01 11:43:41  meichel
** Avoiding to include <stdlib.h> as extern "C" on Borland C++ Builder 4,
**   workaround for bug in compiler header files.
**
** Revision 1.17  1999/06/10 12:12:02  meichel
** Adapted imagectn to new command line option scheme.
**   Added support for Patient/Study Only Q/R model and C-GET (experimental).
**
** Revision 1.16  1999/02/09 14:30:46  meichel
** Made signal handler function parameters compile flag dependent.
**   Now passing (int) unless SIGNAL_HANDLER_WITH_ELLIPSE defined,
**   in which case (...) is passed. Required to avoid conflict between
**   Sun CC 4.2 on Solaris and SGI C++ 4.0 on IRIX.
**
** Revision 1.15  1998/07/24 14:15:51  meichel
** changed ti's signal handler parameters from (int) to (...)
**   to avoid a warning on IRIX 5.3 using SGI C++ 4.0.
**
** Revision 1.14  1998/02/06 15:07:35  meichel
** Removed many minor problems (name clashes, unreached code)
**   reported by Sun CC4 with "+w" or Sun CC2.
**
** Revision 1.13  1998/01/14 14:29:03  hewett
** Modified existing -u command line option to also disable generation
** of UT and VS (previously just disabled generation of UN).
**
** Revision 1.12  1997/08/05 07:46:33  andreas
** - Change needed version number of WINSOCK to 1.1
**   to support WINDOWS 95
**
** Revision 1.11  1997/07/21 08:59:45  andreas
** - Replace all boolean types (BOOLEAN, CTNBOOLEAN, DICOM_BOOL, BOOL)
**   with one unique boolean type OFBool.
**
** Revision 1.10  1997/06/26 12:55:31  andreas
** - Calling of signal function only for existing signals (Windows NT/95
**   do not have the same set of signals as UNIX).
**
** Revision 1.9  1997/05/29 17:06:35  meichel
** All dcmtk applications now contain a version string
** which is displayed with the command line options ("usage" message)
** and which can be queried in the binary with the "ident" command.
**
** Revision 1.8  1997/05/22 13:31:40  hewett
** Modified the test for presence of a data dictionary to use the
** method DcmDataDictionary::isDictionaryLoaded().
**
** Revision 1.7  1997/03/27 16:11:40  hewett
** Added command line switches allowing generation of UN to
** be disabled (it is enabled by default).
**
** Revision 1.6  1996/09/27 08:46:23  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C.
**
** Revision 1.5  1996/05/30 17:41:27  hewett
** Added command line argument (-DD) to explicitly disable communication
** with ctndisp and override the configuration file.
**
** Revision 1.4  1996/05/06 07:33:14  hewett
** Rearranged handing of display disabling code.
**
** Revision 1.3  1996/04/25 16:29:09  hewett
** Added char* parameter casts to bzero() calls.
**
** Revision 1.2  1996/04/22 16:33:03  hewett
** Removed calls to DSU_setTraceLevel().
**
** Revision 1.1  1996/04/22 10:27:25  hewett
** Initial release.
**
*/
